#!/usr/bin/env -S python
from fastapi import FastAPI
import pydantic
import uvicorn
from starlette.middleware.cors import CORSMiddleware
from starlette.status import HTTP_401_UNAUTHORIZED,HTTP_500_INTERNAL_SERVER_ERROR
from fastapi.security import OAuth2PasswordBearer
from fastapi import Depends, FastAPI, HTTPException
from authlib.integrations.starlette_client import OAuth
from starlette.requests import Request
from starlette.middleware.sessions import SessionMiddleware
from starlette.responses import RedirectResponse
import requests
from typing import List
from sqlalchemy.orm import Session

import os
import user_crud 
import user_model 
import user_schema 
import user_jwt
from database import SessionLocal, engine

user_model.Base.metadata.create_all(bind=engine)

app = FastAPI()

origins = [
    "http://localhost",
    "http://localhost:4200",
    "*"
]

oauth = OAuth()
oauth.register(
    name='github',
    client_id=os.environ.get('SECRET_ghclientid', ''),
    client_secret=os.environ.get('SECRET_ghclientsecret', ''),
    access_token_url='https://github.com/login/oauth/access_token',
    access_token_params=None,
    authorize_url='https://github.com/login/oauth/authorize',
    authorize_params=None,
    api_base_url='https://api.github.com/',
    client_kwargs={'scope': 'user:email, repo'},
)

app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)
app.add_middleware(SessionMiddleware, secret_key = "foo") 

oauth2_scheme = OAuth2PasswordBearer(tokenUrl="/login")

def get_db():
    db = SessionLocal()
    try:
        yield db
    finally:
        db.close()

@app.get("/ready")
async def ready():
    return {"ready": True}


@app.get("/auth/login_github")
async def login_github(request: Request):
    redirect_uri = os.environ['AUTHORIZE_URL']
    return await oauth.github.authorize_redirect(request, redirect_uri)
    
@app.get("/auth/authorize_github")
async def authorize_github(request: Request,  db: Session = Depends(get_db)):
    token = await oauth.github.authorize_access_token(request)
    headers = {'Authorization': 'token ' + token['access_token']}
    res = requests.get('https://api.github.com/user', headers=headers)
    if int(res.status_code/100) != 2:
        raise HTTPException(
        status_code=res.status_code,
        detail='Error getting getting github user information')
    profile = res.json()
    db_user: user_model.User = user_crud.get_user_by_github_username(db, str(profile.get('login','')).lower())
    if not db_user:
        return RedirectResponse("{}?user={}".format(os.environ['AUTH_ERROR_URL'], profile['login']))
    user : user_schema.User = user_schema.User.from_orm(db_user)
    user.external_token = token['access_token']
    
    access_token = user_jwt.get_jwt(user)
    try:
        access_token = access_token.decode()
    except:
        pass
    url = '{}?token={}'.format(os.environ['LOGIN_URL'], str(access_token))
    return RedirectResponse(url)


@app.post("/auth/login")
async def login(params: user_schema.LoginParams, db: Session = Depends(get_db)):
    """
    Login the user with the given params. Returns a jwt that expires in 
    15 mins
    
    params: user.LoginParams username and password
    """
    user: user_model.User = user_crud.login(db, params)
    if not user:
        raise HTTPException(
            status_code=HTTP_401_UNAUTHORIZED,
            detail="Incorrect username or password",
            headers={"WWW-Authenticate": "Bearer"},
        )
    else:
        access_token = user_jwt.get_jwt(user_schema.User.from_orm(user))
        return {"access_token": access_token, "token_type": "bearer"}
        

@app.get("/user", response_model=user_schema.User)
async def get_user_profile(token: str = Depends(oauth2_scheme), db: Session = Depends(get_db)):
    """
    Get the user profile from the token
    """
    try:
        decoded = user_jwt.decode_token(token)
        user = user_crud.get_user_by_email(db, decoded.get('username',''))
        if not user:
            raise HTTPException(
                status_code=HTTP_401_UNAUTHORIZED,
                detail='Invalid token',
                headers={"WWW-Authenticate": "Bearer"},
            )
        user.external_token = decoded.get('external_token', None)
        return user
    except Exception as err:
        print("Error {}".format(err))
        raise HTTPException(
            status_code=HTTP_500_INTERNAL_SERVER_ERROR,
            detail=err,
            headers={"WWW-Authenticate": "Bearer"},
        )

@app.post("/auth/users/", response_model=user_schema.User)
def create_user(user: user_schema.UserCreate, db: Session = Depends(get_db)):
    db_user = user_crud.get_user_by_email(db, email=user.email)
    if db_user:
        raise HTTPException(status_code=400, detail="Email already registered")
    return user_crud.create_user(db=db, user=user)

@app.put("/auth/users/{user_email}/edit")
def edit_user(user_email: str, user: user_schema.UserBase, db: Session = Depends(get_db)):
    db_user:user_model.User = user_crud.get_user_by_email(db, user_email)
    if not db_user:
        raise HTTPException(status_code=404, detail="User not found")

    db_user.email = user.email
    db_user.github_username = user.github_username
    db_user.github_org = user.github_org
    db_user.roles = user.roles
    db.commit()
    return user

@app.delete("/auth/users/{user_email}")
def delete_user(user_email: str, db: Session = Depends(get_db)):
    db_user = user_crud.get_user_by_email(db, email=user_email)
    if not db_user:
        raise HTTPException(status_code=404, detail="user does not exist")
    db.delete(db_user)
    db.commit()

@app.post("/auth/users/{user_email}/reset")
def reset_user(user_email: str, db: Session = Depends(get_db)):
    db_user = user_crud.reset_password(db, email=user_email)
    if not db_user:
        raise HTTPException(status_code=404, detail="user does not exist")

@app.get("/auth/users/", response_model=List[user_schema.User])
def read_users(skip: int = 0, limit: int = 100, db: Session = Depends(get_db)):
    users = user_crud.get_users(db, skip=skip, limit=limit)
    return users

@app.get("/auth/users/{user_id}", response_model=user_schema.User)
def read_user(user_id: int, db: Session = Depends(get_db)):
    db_user = user_crud.get_user(db, user_id=user_id)
    if db_user is None:
        raise HTTPException(status_code=404, detail="User not found")
    return db_user   


    
@app.get("/auth/isadmin")
def is_admin(token: str = Depends(oauth2_scheme), db: Session = Depends(get_db)):
    try:
        decoded = user_jwt.decode_token(token)
        db_user = user_crud.get_user_by_email(db, decoded.get('username', ''))
        if not db_user:
            raise HTTPException(
                status_code=HTTP_401_UNAUTHORIZED,
                detail='Invalid token',
                headers={"WWW-Authenticate": "Bearer"},
            )
       
        user : user_schema.User = user_schema.User.from_orm(db_user)
        roles  = user.roles.split(",") if user.roles else []
        for role in roles:
            if role.strip() == "admin":
                print("returning true")
                return True
        return False
    except Exception as err:
        raise HTTPException(
            status_code=HTTP_500_INTERNAL_SERVER_ERROR,
            detail=err,
            headers={"WWW-Authenticate": "Bearer"},
        )

@app.get("/auth/isgithub")
def is_github(token: str = Depends(oauth2_scheme), db: Session = Depends(get_db)):
    try:
        decoded = user_jwt.decode_token(token)
        db_user = user_crud.get_user_by_email(db, decoded.get('username', ''))
        if not db_user:
            raise HTTPException(
                status_code=HTTP_401_UNAUTHORIZED,
                detail='Invalid token',
                headers={"WWW-Authenticate": "Bearer"},
            )
       
        token = decoded.get('external_token', None)
        return not token
    except Exception as err:
        raise HTTPException(
            status_code=HTTP_500_INTERNAL_SERVER_ERROR,
            detail=err,
            headers={"WWW-Authenticate": "Bearer"},
        )


if __name__ == "__main__":
    user_crud.create_admin(next(get_db()))
    uvicorn.run(app, host="0.0.0.0", port=int(os.environ["AUTH_PORT"]), log_level="info")
