from pydantic import BaseModel
import jwt
import bcrypt
from datetime import datetime, timedelta
import uuid
import psycopg2
from models import db
from typing import Dict,List
import os

TOKEN_EXPIRY=timedelta(minutes=15)
SECRET='change_secret'
ALGORITHM='HS256'


class LoginParams(BaseModel):
    username: str
    password: str


class User(BaseModel):
    username: str
    user_id: str
    hashed_pwd: str
    email: str = None

class UserProfile(BaseModel):
    username: str
    user_id: str
    external_token: str = None


fake_db = {
    'admin':  User(username="admin",
                hashed_pwd=b"$2b$12$Dd0ocBuXQXMy2Jj1lec7A.RPn.3GO75JBRU9voDZwxL6AUa/Z20rK",
                user_id="4450e9c7-d974-49bc-87f4-0e6aee7f7340")
}

def init_user_table():
    print("Initializing user table")
    conn = db.DBConnection.getInstance(
        os.environ["DB_HOST"],
        os.environ["DB_PORT"], 
        os.environ["DB_NAME"],
        os.environ["DB_PASSWORD"],
        os.environ["DB_USER"]
    )
    cur = conn.conn.cursor()
    cur.execute('CREATE TABLE IF NOT EXISTS users (username varchar, user_id varchar, hashed_pwd varchar, email varchar, PRIMARY KEY(user_id))')
    conn.conn.commit()
    
    cur.execute('SELECT * FROM users WHERE username=%s', ('admin',))
    user = cur.fetchone()
    if not user:   
        add_db_user(fake_db.get("admin"))
    else:
        print("User already created")
    
def fetch_db_user(name: str):
    for retry in range(3):
        try:
            conn = db.DBConnection.getInstance(
                os.environ["DB_HOST"],
                os.environ["DB_PORT"], 
                os.environ["DB_NAME"],
                os.environ["DB_PASSWORD"],
                os.environ["DB_USER"], reconnect=True)
            cur = conn.conn.cursor()
            cur.execute('SELECT * FROM users WHERE username=%s', (name,))
            user = cur.fetchone()
            if user:
                return User(user_id= user[1], username= user[0], hashed_pwd=user[2], email= user[3])
        except Exception as err:
            print("Error fetching data {}", err)
   
    return None

def add_db_user(user: User):
    conn = db.DBConnection.getInstance(
        os.environ["DB_HOST"],
        os.environ["DB_PORT"], 
        os.environ["DB_NAME"],
        os.environ["DB_PASSWORD"],
        os.environ["DB_USER"]
    )
    
    cur = conn.conn.cursor()
    cur.execute('INSERT INTO users (user_id, username, hashed_pwd, email) VALUES(%s,%s,%s,%s)'
    ,(user.user_id,user.username, user.hashed_pwd,user.email))
    conn.conn.commit()
   
    
def fetch_user(params: LoginParams):
    print("Fetching User {}".format(params.username));
    user = fetch_db_user(params.username)
    if not user:
        print("User {} missing from database".format(params.username))
        return None
    if bcrypt.checkpw(params.password.encode(), user.hashed_pwd.encode()):
        return user
    return None

def fetch_github_user(profile: Dict[str,str]):
    #TODO: remove this once organizations exit
    if profile['login'].lower() not in ['muchene', 'ssakhuja57', 'dhruvbs', 'aramse', 'siyuanma13', 'edgerunnerio', 'sk-Predica']:
        return None
    username = '{}@githhub.com'.format(profile['login'])
    user = fetch_db_user(username)
    if not user:
        user = User(username=username, hashed_pwd='external', user_id = str(uuid.uuid4()))
        add_db_user(user)
        return user
    return user

def refresh_token(token: str):
    decoded = None
    try:
        decoded = jwt.decode(token,SECRET,algorithm=[ALGORITHM])
    except jwt.ExpiredSignatureError as err:
        decoded = jwt.decode(token, SECRET, algorithm=[ALGORITHM], verify=False)
    decoded['exp'] = datetime.now()+ TOKEN_EXPIRY
    return jwt.encode(decoded, SECRET, algorithm=ALGORITHM)

def get_user_profile(token: str):
    print("Getting user profile")
    decoded = None
    try:
        decoded = jwt.decode(token,SECRET,algorithm=[ALGORITHM])
    except jwt.ExpiredSignatureError as err:
        decoded = jwt.decode(token, SECRET, algorithm=[ALGORITHM], verify=False)
    user = fetch_db_user(decoded['username'])
    if user:
        return UserProfile(username=user.username, user_id=user.user_id, external_token=decoded.get('external_token', ''))
    return None

def get_jwt(user: User, external_token:str = ''):
    return jwt.encode({'username': user.username, 'exp': datetime.now()+TOKEN_EXPIRY, 'external_token': external_token}, SECRET, algorithm=ALGORITHM)
