import { Component, OnInit, Inject, OnDestroy} from '@angular/core';
import { MatDialog, MatDialogRef, MAT_DIALOG_DATA } from '@angular/material/dialog';
import { ApplicationsService } from '../../services/applications.service';
import { GeneralService } from '../../services/general.service';
import { MatTableDataSource } from  '@angular/material/table';
import { Globals } from '../../global';
import { takeUntil } from 'rxjs/operators';
import { Subject } from 'rxjs';

@Component({
  selector: 'app-dialog',
  templateUrl: './app-dialog.component.html',
  styleUrls: ['./app-dialog.component.css']
})
export class AppDialogComponent implements OnInit, OnDestroy {
  protected ngUnsubscribe: Subject<void> = new Subject<void>();
  
  loading=0;
  errors={};

  name;
  servingEnv;
  repo;
  description:string;
  envData;
  appData;
  selectedEnv;
  selectedEnvDetailed;
  selected:string; //used for environment selection 

  environmentSource = new MatTableDataSource([]);
  environmentDisplayedColumns:string[] = ['name', 'type', 'cluster', 'version', 'created', 'updated', 'details'];
  
  architectureSource = new MatTableDataSource([]);
  architectureDisplayedColumns:string[] = ['component', 'type', 'instances', 'url'];

  architectureSource2 = new MatTableDataSource([]);
  architectureDisplayedColumns2:string[] = ['test', 'instances'];



  constructor(
    private applicationsService: ApplicationsService,
    private generalService: GeneralService,
    public dialogRef: MatDialogRef<AppDialogComponent>,
    @Inject(MAT_DIALOG_DATA) public data: any,
    private global:Globals) {}

  onNoClick(): void {
    this.dialogRef.close();
  }

  ngOnInit(): void {
    this.initialize();
  }

  public ngOnDestroy(): void {
    // This aborts all HTTP requests.
    this.ngUnsubscribe.next();
    // This completes the subject properlly.
    this.ngUnsubscribe.complete();
  }

  initialize(){

    this.loading++;
    this.name = this.data.name;
    this.applicationsService.getAppEnvs(this.data.name)
    .pipe( takeUntil(this.ngUnsubscribe) )
    .subscribe(
      (data) => {
        this.envData=data['results'];
        this.servingEnv = data['meta']['serving_env'];
        //populate Environments Table
        this.populateEnvironmentsTable();
        //set default env name
        this.selectedEnv = this.envData[0];
        for (var i in this.envData){
          var env =this.envData[i];
          if (env['name'].startsWith('prod') || env['type'].startsWith('prod') || env['name'] == 'master' || env['name'] == 'main'){
            this.selectedEnv = env;
          }
        }

        this.selected=this.selectedEnv.name;
        //second API Call to populate description, architecture tables, and environment dropdown 
        this.applicationsService.getAppDetails(this.data.name, this.selected)
        .pipe( takeUntil(this.ngUnsubscribe) )
        .subscribe(
          (data2) => {
           
            this.appData = data2['results'];
            
            //set detailed env 
            this.selectedEnv = this.appData[0];
            this.repo = this.selectedEnv.repo;
            this.selected=this.selectedEnv.name
          
            //Populate Description
            if (this.selectedEnv.cfg.general['description']){
              this.description = ' -  '+this.selectedEnv.cfg.general.description;
            }
            //Populate Architecture Table
            this.populateArchitectureTables();
            this.loading--;
            if (this.errors['general']){
              delete this.errors['general'];
            }
          },
          (error) => {

            this.errors['general']=this.global.ORCH_CONNECTION_ERROR;
            this.loading--;
          }
        );
        if (this.errors['general']){
          delete this.errors['general'];
        };
      },
      (error) => {

        this.errors['general']=this.global.ORCH_CONNECTION_ERROR;
        this.loading--;
      }
      
    )
  }

  populateEnvironmentsTable(){
    var environmentData=[];
    for (var i in this.envData){

      var env = this.envData[i];
      var environmentName = env['name'];
      var environmentType = env['type'];
      var environmentCluster = env['cluster'];
      var environmentVersion = env['tag'];
      var environmentCreated = this.generalService.timeSince(env['created']);
      var environmentUpdated = this.generalService.timeSince(env['updated']);
      var environmentDetails = 'environments?app='+this.name+'&env='+environmentName+'&cluster='+environmentCluster;
      var environmentForSorting = env['updated'];
      var environmentObject = {};
      environmentObject['name']=environmentName;
      environmentObject['type']=environmentType;
      environmentObject['cluster']=environmentCluster;
      environmentObject['version']=environmentVersion;
      environmentObject['created']=environmentCreated;
      environmentObject['updated']=environmentUpdated;
      environmentObject['details']=environmentDetails;
      environmentObject['forSorting'] = environmentForSorting;
      
      environmentData.push(environmentObject);
    }
    environmentData.sort((a,b) => {
      if (a.forSorting==null){
        return 1; 
      } else if (b.forSorting==null){
        return -1;
      } else if (a.forSorting < b.forSorting){
        return 1;
      } else if (a.forSorting >= b.forSorting){
        return -1;
      }
    });
    this.environmentSource.data=environmentData;
  }


  populateArchitectureTables(){
    var architectureData=[];
    for (var i in this.selectedEnv.components){
      var component = this.selectedEnv.components[i];
      var architectureComponent = component['name'];
      var architectureType = component.kind.split('-')[0]
      var architectureInstances = component.instances.length;
      var architectureURL=[];
      if (component.urls){
        for (var i in component.urls){
          architectureURL.push(component.urls[i]);
        }
      } 
      var architectureObject={};
      architectureObject['component'] = architectureComponent;
      architectureObject['type'] = architectureType;
      architectureObject['instances'] = architectureInstances;
      architectureObject['url'] = architectureURL;
      architectureData.push(architectureObject);
    }
    this.architectureSource.data=architectureData;

    //Architecture Table 2
    var architectureData2=[];
    for (var i in this.selectedEnv.tests){
      var test = this.selectedEnv.tests[i];
      var architectureTest = test.name;
      var architectureInstances2 = test.instances.length;
      var architectureObject2={};
      architectureObject2['test'] = architectureTest;
      architectureObject2['instances'] = architectureInstances2;
      architectureData2.push(architectureObject2);
    }
    this.architectureSource2.data = architectureData2;
  }

  updateArchitectureTables(environmentName){
    this.loading++;
    this.applicationsService.getAppDetails(this.data.name, environmentName)
    .pipe( takeUntil(this.ngUnsubscribe) )
    .subscribe(
      (data2) => {
        this.appData = data2['results'];

        //set detailed env 
        this.selectedEnv = this.appData[0];
        this.repo = this.selectedEnv.repo;
        this.selected=this.selectedEnv.name
      
        //Populate Description
        this.description = this.selectedEnv.cfg.general.description;

        //Populate Architecture Table
        this.populateArchitectureTables();

        this.loading--;
        if (this.errors['general']){
          delete this.errors['general'];
        }
      },
      (error) => {

        this.errors['general']=this.global.ORCH_CONNECTION_ERROR;
        this.loading--;
      }
    )
  }

}
