import { ChangeDetectorRef, Component, OnInit, OnDestroy } from '@angular/core';
import { MediaMatcher } from '@angular/cdk/layout';
import { MatDialog, MatDialogRef, MAT_DIALOG_DATA } from '@angular/material/dialog';
import { Globals } from '../global'
import { GeneralService } from '../services/general.service';
import { ConfigBuilderService } from '../services/config-builder.service';
import { CompTemplatesService } from '../../app/services/comp-templates.service';
import { first, take } from 'rxjs/operators';

import { InfoDialogComponent } from '../info-dialog/info-dialog.component';
import { InitDialogComponent } from './init-dialog/init-dialog.component';
import { FormGroup, FormControl, Validators, FormBuilder } from '@angular/forms';
import { Subscription, Subject } from 'rxjs';
import { ActivatedRoute } from '@angular/router';
import { takeUntil } from 'rxjs/operators';
import { Title } from '@angular/platform-browser';
import { CommitDialogComponent } from '../commit-dialog/commit-dialog.component';
import { GithubService } from '../services/github.service';
import { HttpErrorResponse } from '@angular/common/http';
import { ConfirmDialogComponent } from '../confirm-dialog/confirm-dialog.component';

interface Repository {
  name: string;
  owner: string;
}


@Component({
  selector: 'app-config-builder',
  templateUrl: './config-builder.component.html',
  styleUrls: ['./config-builder.component.css']
})
export class ConfigBuilderComponent implements OnInit, OnDestroy {


  protected ngUnsubscribe: Subject<void> = new Subject<void>();

  infoDialog: string =
    "<p>" +
    "This page can be used to help build/modify f8.yaml configuration for any application." +
    "<ul>" +
    "<li>The interactive configuration builder is integrated with the core f8.yaml schema, so it includes all available fields.</li>" +
    "<li>The builder and the text area mirror each other (updates to one are automatically reflected in the other).</li>" +
    "<li>Hover over any button for a description of its function.</li>" +
    "</ul>" +
    "</p>";

  mobileQuery: MediaQueryList;

  errors = '';
  loading = 0;

  //query parameters
  routeQueryParams$: Subscription;
  queryParamSourceRepoURL;
  queryParamSourceRepoRef;
  queryParamAppName;

  //for displaying
  show_env_type_resolver: boolean = false;
  show_field_help: boolean = true;

  //for jsoneditor library
  schema;

  //elements
  builder_form;
  raw_config;
  resolved_env_config;

  //resolved_env_config 
  init_success_message = '';

  //formgroup for resolver
  resolverForm = new FormGroup({
    env_type_name: new FormControl('local', [Validators.required]),
    include_globals: new FormControl()
  })


  public repoForm: FormGroup;
  public repos: Repository[];

  //public selectedRepo: FormControl;

  public branches: string[];
  //public selectedBranch: FormControl;

  private defaultRepo: Repository;
  private defaultBranch
  get env_type_name(): any {
    return this.resolverForm.get('env_type_name');
  }
  get include_globals(): any {
    return this.resolverForm.get('include_globals');
  }


  private _mobileQueryListener: () => void;

  constructor(
    changeDetectorRef: ChangeDetectorRef,
    media: MediaMatcher,
    public dialog: MatDialog,
    private global: Globals,
    public generalService: GeneralService,
    private configBuilderService: ConfigBuilderService,
    private compTemplatesService: CompTemplatesService,
    private route: ActivatedRoute,
    private titleService: Title,
    private githubService: GithubService,
    private formBuilder: FormBuilder) {


    titleService.setTitle('Config Builder');
    this.mobileQuery = media.matchMedia('(max-width: 600px)');
    this._mobileQueryListener = () => changeDetectorRef.detectChanges();
    this.mobileQuery.addListener(this._mobileQueryListener);

  }

  public ngOnDestroy(): void {
    this.mobileQuery.removeListener(this._mobileQueryListener);
    //this.routeQueryParams$.unsubscribe();
    // This aborts all HTTP requests.
    this.ngUnsubscribe.next();
    // This completes the subject properly.
    this.ngUnsubscribe.complete();
  }

  async ngOnInit() {
    this.loading++;
    await this.createRepoForm();
    this.resolved_env_config = document.getElementById('resolved-env-config');
    this.raw_config = document.getElementById('raw-config');
    this.builder_form;
    this.configBuilderService.getSchemas()
      .pipe(takeUntil(this.ngUnsubscribe))
      .subscribe(
        (data) => {
          this.schema = data;

          this.builder_form = this.generalService.create_form_config_builder('builder-form', this.schema, {}, undefined);

          this.builder_form.on('change', () => {

            // update raw yaml
            this.raw_config.value = this.generalService.YAML.stringify(this.builder_form.getValue(), 20, 2);
            this.generalService.set_field_help(this.show_field_help);

            // callback to update vars schema
            const updateVarsSchema = function (comp_path) {
                console.log('updates at watched field for component at path: ' + comp_path + ', updating vars schema');
                var vars_path = comp_path + '.vars';
                var template_path = comp_path + '.template';
                var vars_editor = this.builder_form.getEditor(vars_path);
                var template_name = this.builder_form.getEditor(template_path).getValue();
                if (template_name == '' || vars_editor == null)
                  return;

                this.loading++;

                vars_editor.cached_editors = {};
                for (let i in vars_editor.editors)
                  vars_editor.removeObjectProperty(i);
                vars_editor.disable();
                this.compTemplatesService.getCompTemplateVarsSchema(template_name)
                  .subscribe(
                    (data) => {
                      vars_editor.schema = data;
                      var current_vars = vars_editor.getValue();
                      var initial = {};
                      for (let k in vars_editor.schema.required)
                        initial[vars_editor.schema.required[k]] = null;
                      vars_editor.setValue(initial);
                      vars_editor.enable();
                      this.loading--;
                    },
                    (error) => {
                      vars_editor.enable();
                      this.loading--;
                      this.errors = 'error loading variables for template: ' + template_name;
                    }
                  );
            }

            var comps = this.builder_form.getEditor('root.components');
            for (let i in comps.value) {
              var comp = comps.value[i];
              var comp_path = 'root.components.' + i;
              var comp_editor = this.builder_form.getEditor(comp_path);

              //this.builder_form.getEditor(comp_path + '.type').disable();  // this doesn't work consistently

              if (comp.hasOwnProperty('template')) {
                var template_path =  comp_path + '.template';
                var vars_path = comp_path + '.vars';

                // updateVarsSchema() on changes to "template" field
                this.builder_form.unwatch(template_path);
                this.builder_form.watch(template_path, updateVarsSchema.bind(this, comp_path));

                // validate on changes to "vars" (broken currently)
                //this.builder_form.unwatch(vars_path);
                //this.builder_form.watch(vars_path, function(){  // <- ideally would just call update_form but this causes endless loop
                //    this.errors = 'template variables updated, click in/out of above window to validate';
                //}.bind(this));
              }
            }
            if (this.errors != '' || this.builder_form.validate().length)
              this.update_form(null, false);
          });

          this.loading--;
          this.handleQueryParameters();
        },
        (error) => {
          this.errors = 'Error trying to get configuration schema';
          this.loading--;
        }
      );
  }

  async createRepoForm() {
    await this.fetchRepos();
    const params = await this.route.queryParams.pipe(first()).toPromise();
    let selectedRepo = '';
    let selectedBranch = '';

    if (params['repo_url']) {
      this.queryParamSourceRepoURL = decodeURIComponent(params['repo_url']);
    }

    if (params['repo_ref']) {
      this.queryParamSourceRepoRef = params['repo_ref']
    } else {
      this.queryParamSourceRepoRef = 'master';
    }
    if(params['app']) {
      this.queryParamAppName = params['app']
    }

    if (this.queryParamSourceRepoURL) {
      const splits = this.queryParamSourceRepoURL.split('/');
      const owner = splits[3];
      const repo_name = splits[4].split(".git")[0];
      this.defaultRepo = this.repos.find(repo => {
        return repo.owner === owner && repo.name === repo_name;
      });
      if(this.defaultRepo) {
        await this.fetchBranches(this.defaultRepo.name);
        this.defaultBranch = this.branches.find(branch =>{ return branch == this.queryParamSourceRepoRef});
      }
    }
    
    this.repoForm = this.formBuilder.group({
      selectedRepo: [this.defaultRepo, Validators.required],
      selectedBranch: [this.defaultBranch , Validators.required]
    });

  }

  async fetchRepos() {
    //TODO: make api call.
    this.repos = [];
    try {
      const data: any[] = await this.githubService.getRepos().toPromise();
      if(data.length == 0){
        return;
      }
      data.forEach(element => {
        this.repos.push({
          owner: element['owner']['login'],
          name: element['name']
        });
      });
    } catch (error) {
      this.errors = "Unable to fetch repos, are you logged in with GitHub?";
    }

  }

  async fetchBranches(repo_name: string) {
    repo_name = repo_name ? repo_name : this.repoForm.controls.selectedRepo.value.name;
    this.branches = [];
    try {
      const data : any[] = await this.githubService.getBranches(repo_name).toPromise();
      data.forEach(branch => this.branches.push(branch['name']));
    }catch (error) {
      this.errors = "Unable to fetch branches for repo";
    }
  }


  handleQueryParameters() {
    if (this.queryParamSourceRepoURL) {

      this.loading++; 
      this.configBuilderService.initConfig(this.queryParamSourceRepoURL, this.queryParamSourceRepoRef, this.queryParamAppName)
        .pipe(takeUntil(this.ngUnsubscribe))
        .subscribe(
          (data) => {
            if (data['errors'].length) {
              this.errors = data['errors'].join("\n-> ");
            } else {
              this.raw_config.value = data['result'];
              // this.init_config_value = data['result'];
              // this.errors='';
              // this.dialogRef.close(
              //   {
              //     data: this.init_config_value,
              //     env: source_repo_ref,
              //     repoURL: source_repo_url
              //   }
              // );
              this.update_form(null, false);
            }
            this.loading--;
          },
          (error) => {
            this.errors = 'Error in trying to initialize config';
            this.loading--;
          }
        );
    }
  }

  openInfoDialog(): void {
    const dialogRef = this.dialog.open(InfoDialogComponent,
      {
        width: '60%',
        height: '75%',
        data: {
          type: "config-builder",
          text: this.infoDialog
        }
      }
    );
  }

  toggle_field_help(): void {
    this.show_field_help = !this.show_field_help;
    this.generalService.set_field_help(this.show_field_help);
  }

  save_config() {
    this.update_form(null, true)
  };

  openInitDialog() {
    this.init_success_message = '';
    const dialogRef = this.dialog.open(InitDialogComponent,
      {
        width: '60%',
        height: '75%',
        data: {
          type: "config-builder"
        }
      }
    );

    dialogRef.afterClosed().subscribe(
      (result) => {
        this.raw_config.value = result['data'];
        this.update_form(null, false);
        // this.init_success_message = "successfully initialized configuration";
      }
    );
  }

  uploadGithub() {

    const dialogRef = this.dialog.open(ConfirmDialogComponent,
      {
        width: '60%',
        height: '40%',
        data: {
          title: 'Confirm Upload',
          message: `Are you sure you want to upload the f8.yaml file to repo ${this.repoForm.controls.selectedRepo.value.name}, branch: ${this.repoForm.controls.selectedBranch.value} ?`
        }
      }
    );

    dialogRef.afterClosed().subscribe(
      (result: boolean) => {
        if (result) {
          this.loading++;
          this.githubService
            .commitFile(this.repoForm.controls.selectedRepo.value.name, this.repoForm.controls.selectedBranch.value, "f8.yaml", this.raw_config.value)
            .subscribe(data => { this.loading-- }, (err: HttpErrorResponse) => { this.errors = err.statusText; this.loading-- });
        }
      }
    )

  }

  downloadF8() {
    const dialogRef = this.dialog.open(ConfirmDialogComponent,
      {
        width: '60%',
        height: '40%',
        data: {
          title: 'Confirm Download',
          message: `Are you sure you want to load the f8.yaml file from repo ${this.repoForm.controls.selectedRepo.value.name}, branch: ${this.repoForm.controls.selectedBranch.value} ?`
        }
      }
    );

    dialogRef.afterClosed().subscribe(
      (result: boolean) => {
        if (result) {
          this.loading++;
          this.githubService
            .getFile(this.repoForm.controls.selectedRepo.value.name, this.repoForm.controls.selectedBranch.value, 'f8.yaml')
            .subscribe(
              (data) => {
                this.raw_config.value = atob(data['content']);
                this.update_form(null, false);
                this.loading--;
              },
              (err: HttpErrorResponse) => { this.errors = err.statusText; this.loading--; }
            );
        }
      }
    )

  }

  openRepoLink() {
    const selRepo = this.repos.filter(repo => {
      return repo.name === this.repoForm.controls.selectedRepo.value.name;
    });
    const owner = selRepo[0].owner;

    let link = `https://github.com/${owner}/${this.repoForm.controls.selectedRepo.value.name}/tree/${this.repoForm.controls.selectedBranch.value}`;
    window.open(link);
  }


  update_form(callback, toSave: boolean): string {
    if (this.hasOwnProperty('resolved_env_config'))
      this.resolved_env_config.value = 'Base configuration changed, click Resolve to re-calculate (after fixing any reported errors)';
    this.init_success_message = '';
    this.errors = '';

    try {
      var config_obj = this.generalService.YAML.parse(this.raw_config.value);
    } catch (e) {
      this.errors = e.message;
      return;
    }

    this.loading++;
    this.builder_form.disable();
    this.init_success_message = 'validating configuration ...';
    this.configBuilderService.validateWithServer(this.raw_config.value).
      subscribe(
        (data) => {

          if (data['result'] === 'successful') {

            this.errors = '';
            this.init_success_message = 'configuration is valid';
            this.builder_form.setValue(config_obj);
            if (callback !== null) {
              callback();
            }
            if (toSave) {
              this.generalService.saveTextAsFile(this.raw_config.value, "f8.yaml");
            }
          } else {
            this.init_success_message = '';
            this.errors = data['errors'].join("\n-> ");
          }
          this.builder_form.enable();
          this.loading--;
          return this.errors;
        },
        (error) => {
          this.init_success_message = '';
          this.errors = this.global.ORCH_CONNECTION_ERROR;
          this.builder_form.enable();
          this.loading--;
          return this.errors;
        }
      );

  };

  resolve_env_type(env_type_name, include_globals, raw_config_value) {
    this.update_form(() => {
      if (this.errors.length) {
        return;
      }
      this.loading++;
      this.configBuilderService.resolveEnvironmentType(env_type_name, include_globals, raw_config_value).
        subscribe(
          (data) => {

            if (data['errors'].length) {
              this.errors = data['errors'].join("\n-> ");
            } else {
              this.resolved_env_config.value = data['result'];
            }
            this.loading--;
          },
          (error) => {
            this.errors = this.global.ORCH_CONNECTION_ERROR;
            this.loading--;

          }
        );
    }, false);
  };


}
